/**
 * @file Main.cpp
 * @date 2015-04-19
 * @brief Tv\[XR[h
 */

#include "HamFramework\SceneManager.hpp"

#define FSL L"Large"
#define FSM L"Medium"
#define FSS L"Small"

#define AREA_HEIGHT (80)

/// <summary>1ʕ</summary>
struct EnvFace {
	/// <summary>]</summary>
	int mirror = 0;
	/// <summary>]</summary>
	int rot90s = 0;

	/// <summary>[h摜</summary>
	Image loadImage;
	/// <summary>␳摜</summary>
	Image img;
	/// <summary>eNX`</summary>
	Texture tex;

	/// <summary>90x]</summary>
	void Rotate() {
		++this->rot90s;

		this->update();
	}
	/// <summary>]</summary>
	void Reverse() {
		++this->mirror;

		this->update();
	}

	/// <summary>␳摜XV</summary>
	void update() {
		this->img = this->loadImage.clone();
		if (this->mirror & 1) {
			this->img.mirror();
		}
		for (int i = 0; i < (this->rot90s & 3); ++i) {
			this->img.rotate90();
		}

		this->tex = Texture(this->img);
	}
};

/// <summary>ŜŎgp</summary>
struct Data {
	/// <summary>GUIz</summary>
	Array<GUI> guis;

	/// <summary></summary>
	Array<EnvFace> envFaces;

	/// <summary>bVf[^</summary>
	MeshData meshData;
	/// <summary>bV</summary>
	DynamicMesh mesh;

	/// <summary>}EXʒu</summary>
	Optional<Point> prePos;

	/// <summary>J</summary>
	Vec3 basePYR = { 0, 0, 0 };
	/// <summary>ZoJ</summary>
	Vec3 curPYR;

	/// <summary>bZ[Wp</summary>
	Effect effect;
};

using MyApp = ham::SceneManager<String,Data>;

/// <summary>ʂ̕</summary>
String names[] = { L"", L"", L"", L"", L"k", L"" };

/// <summary>bZ[W\</summary>
struct Message : IEffect {
	String msg;
	int my = 100;

	Message(String inMsg) {
		msg = inMsg;
	}
	/// <summary>XV</summary>
	bool update(double t) override {
		double alpha = 1.0;
		if (t < 1.0) {
			alpha = t;
		}
		else if (t > 4.0) {
			alpha = 5.0 - t;
		}
		ColorF c0(Palette::White, alpha);
		ColorF c1(Palette::Black, alpha);

		int y = my - (int)Math::Floor(t * 20);

		FontAsset(FSS)(msg).drawCenter(y + 2, c1);
		FontAsset(FSS)(msg).drawCenter(y, c0);
		return t < 5.0;
	}
};

/// <summary>CV[</summary>
struct MainScene : MyApp::Scene {

	/// <summary></summary>
	void init() override {
		Window::SetTitle(L"}bv쐬 - [F1]`[F6]ŋ]");

		Window::SetStyle(WindowStyle::Sizeable);
		Dragdrop::SetEnabled(true);

		Graphics3D::SetAmbientLight({0.5,0.5,0.5});

		makeMeshData();
		m_data->mesh = DynamicMesh(m_data->meshData);

		for (int i = 0; i < 6; ++i) {
			EnvFace ef;
			ColorF cs[] = { Palette::Silver, Palette::Gray,
				Palette::Darkseagreen, Palette::Lightgreen,
				Palette::Cornflowerblue, Palette::Cadetblue };

			Image img(512, 512, cs[i]);
			FontAsset(FSL)(names[i]).writeCenter(img,Point(256, 256));
			ef.loadImage = img;
			ef.update();
			m_data->envFaces.push_back(ef);
		}

		GUI gui(GUIStyle::Default);
		gui.setTitle(L"j[");

		gui.add(L"sltwo", GUISlider::Create(5, 12, 10));
		gui.addln(L"txpx",GUIText::Create(L"1024"));

		gui.style.width = 200;

		gui.addln(L"btmake", GUIButton::Create(L".ddso"));

		const Rect rc = gui.getRect();
		gui.setPos(Window::Width() - rc.w - 4, Window::Height() - rc.h - 4);

		m_data->guis.push_back(gui);
	}

	/// <summary>XV</summary>
	void update() override {
		auto nowPos = Mouse::Pos();
		auto ray = Mouse::Ray();

		if (m_data->guis[0].slider(L"sltwo").hasChanged) {
			m_data->guis[0].text(L"txpx").text = Format(getOutPx());
		}

		if (Input::KeyF1.clicked) {
			m_data->envFaces[0].Reverse();
		}
		if (Input::KeyF2.clicked) {
			m_data->envFaces[1].Reverse();
		}
		if (Input::KeyF3.clicked) {
			m_data->envFaces[2].Reverse();
		}
		if (Input::KeyF4.clicked) {
			m_data->envFaces[3].Reverse();
		}
		if (Input::KeyF5.clicked) {
			m_data->envFaces[4].Reverse();
		}
		if (Input::KeyF6.clicked) {
			m_data->envFaces[5].Reverse();
		}

		if (Input::MouseL.pressed) {

			if (m_data->prePos) { // ÂĉĂƂ
				// 𔽉f
				auto prePos = m_data->prePos.value();
				m_data->curPYR.y = m_data->basePYR.y - (nowPos.x - prePos.x) * 0.01;
				m_data->curPYR.x = m_data->basePYR.x + (nowPos.y - prePos.y) * 0.01;
				m_data->curPYR.x = Clamp<double>(m_data->curPYR.x, -HalfPi, HalfPi);
			}
			else { // VƂ
				m_data->prePos.emplace(nowPos);

				if (nowPos.x < AREA_HEIGHT) {
					int index = nowPos.y / AREA_HEIGHT;
					if (index >= 0 && index < 6) {
						m_data->envFaces[index].Rotate();
					}
				}
			}

		}
		else {
			if (m_data->prePos) { // Ƃ
				m_data->basePYR = m_data->curPYR;
				m_data->prePos.reset();
			}
			else { // NbNĂȂƂ
			}
		}

		if (Dragdrop::HasItems()) {
			auto files = Dragdrop::GetFilePaths();
			auto pos = Dragdrop::GetPos();

			int index = pos.y / AREA_HEIGHT;
			if (index >= 0 && index < 6) {
				m_data->envFaces[index].loadImage = Image(files[0]);
				m_data->envFaces[index].update();
			}
		}

		if (m_data->guis[0].button(L"btmake").pushed) {
			makeDDS();
		}

		m_data->effect.update();
	}

	/// <summary>`</summary>
	void draw() const override {
		// sb`ĕp]
		Array<Vec2> fs = { {0,0},
			{HalfPi,0}, {HalfPi,HalfPi}, {HalfPi,Pi}, {HalfPi,HalfPi*3.0},
			{Pi,0} };

		double dist = 30.0;
		Camera cam;
		cam.lookat = { 0, 0, 0 };
		cam.pos = { 0, 0, -dist };
		cam.up = { 0, 1, 0 };
		Vec3 v = { 0, dist * Sin(m_data->curPYR.x), dist * Cos(m_data->curPYR.x) };
		cam.pos = { v.z * Cos(m_data->curPYR.y), v.y, v.z * Sin(m_data->curPYR.y) };
		Graphics3D::SetCamera(cam);

		int twoSize = 64;
		int x = AREA_HEIGHT;
		for (int i = 0; i < 6; ++i) {
			int y = i * AREA_HEIGHT + AREA_HEIGHT/2;
			RoundRect({ x/2 - twoSize / 2, y - twoSize / 2 }, { twoSize, twoSize }, 4).drawFrame(0, 4, Palette::Orangered);
			m_data->envFaces[i].tex.resize({ twoSize, twoSize }).drawAt({ x / 2, y });

			FontAsset(FSS)(names[i]).drawCenter({ x*3/2 - 2, y -2 },
				Palette::Black);
			FontAsset(FSS)(names[i]).drawCenter({ x*3/2,y });
		}

		for (int i = 0; i < 2; ++i) {
			for (int j = 0; j < 6; ++j) {
				int offset = (i*6+j) * 3*2;
				m_data->mesh.drawSubset(offset,3*2,m_data->envFaces[j].tex);
			}
		}

	}

	/// <summary>GUIo̓sNZ擾</summary>
	int getOutPx() const {
		int power = m_data->guis[0].slider(L"sltwo").valueInt;
		return (1 << power);
	}

	/// <summary>OƓ̔</summary>
	void makeMeshData() {
		m_data->meshData.vertices.clear();
		m_data->meshData.indices.clear();

		Vec3 ns[] = {
				{1,0,0}, {-1,0,0},
				{0,1,0}, {0,-1,0},
				{0,0,1}, {0,0,-1}
		};

		Vec3 vs[] = {
				{1,1,-1}, {1,1,1}, {1,-1,-1}, {1,-1,1},
				{-1,1,1}, {-1,1,-1}, {-1,-1,1}, {-1,-1,-1},
				{1,1,-1}, {-1,1, -1}, {1,1,1}, {-1,1,1},
				{1,-1,1}, { -1,-1,1}, {1,-1, -1}, {-1,-1, -1},
				{1,1,1}, {-1,1,1}, {1,-1,1}, {-1,-1,1},
				{-1,1,-1}, {1,1,-1}, {-1,-1,-1}, {1,-1,-1}
		};

		for (int i = 0; i < 2; ++i) {
			double nm = (i == 0) ? 1 : -1;
			double r = (i == 0) ? 2 : 20;
			for (int j = 0; j < 6; ++j) {
				Vec3 n = ns[j] * nm;

				MeshVertex v0;
				v0.position = vs[j*4+0] * r;
				v0.normal = n;
				v0.texcoord = {1,0};
				m_data->meshData.vertices.push_back(v0);
				MeshVertex v1;
				v1.position = vs[j * 4 + 1] * r;
				v1.normal = n;
				v1.texcoord = {0, 0 };
				m_data->meshData.vertices.push_back(v1);
				MeshVertex v2;
				v2.position = vs[j * 4 + 2] * r;
				v2.normal = n;
				v2.texcoord = {1, 1 };
				m_data->meshData.vertices.push_back(v2);
				MeshVertex v3;
				v3.position = vs[j * 4 + 3] * r;
				v3.normal = n;
				v3.texcoord = {0, 1 };
				m_data->meshData.vertices.push_back(v3);

				int offset = (i*6 + j) * 4;

				if (i == 0) {
					m_data->meshData.indices.push_back(offset + 0);
					m_data->meshData.indices.push_back(offset + 1);
					m_data->meshData.indices.push_back(offset + 2);

					m_data->meshData.indices.push_back(offset + 2);
					m_data->meshData.indices.push_back(offset + 1);
					m_data->meshData.indices.push_back(offset + 3);
				}
				else {
					m_data->meshData.indices.push_back(offset + 2);
					m_data->meshData.indices.push_back(offset + 1);
					m_data->meshData.indices.push_back(offset + 0);

					m_data->meshData.indices.push_back(offset + 3);
					m_data->meshData.indices.push_back(offset + 1);
					m_data->meshData.indices.push_back(offset + 2);
				}
			}


		}


	}

	/// <summary>t@C쐬</summary>
	bool makeDDS() const {
		auto path = Dialog::GetSave({ { L"}bv(*.dds)", L"*.dds" } });
		if (path == false) {
			return false;
		}

		auto w = BinaryWriter(path.value());
		int width = getOutPx();
		int height = width;
		int headByte = 128; // DDSHEAD
		int bodyByte = (width / 4) * (height / 4) * 8;
		{
			w.write<int>(0x20534444);
			w.write<int>(124);
			w.write<int>(0);
			w.write<int>(height);
			w.write<int>(width);
			w.write<int>(0);
			w.write<int>(0);
			w.write<int>(0);

			w.write<int>(0);
			w.write<int>(0);
			w.write<int>(0);
			w.write<int>(0);
			w.write<int>(0);
			w.write<int>(0);
			w.write<int>(0);
			w.write<int>(0);

			w.write<int>(0);
			w.write<int>(0);
			w.write<int>(0);
			w.write<int>(32);
			w.write<int>(4);
			w.write<int>(0x31545844); // DXT1
			w.write<int>(0);
			w.write<int>(0);

			w.write<int>(0);
			w.write<int>(0);
			w.write<int>(0);
			w.write<int>(0x00001008);
			w.write<int>(0x0000fe00);
			w.write<int>(0);
			w.write<int>(0);
			w.write<int>(0);
		}

		try {
			std::unique_ptr<char[]> buf = std::make_unique<char[]>(bodyByte);

			for (int i = 0; i < 6; ++i) {
				Image img = m_data->envFaces[i].img;
				img.scale({ width, height });
				auto one = img.encodeDDS(DDSFormat::DXT1);

				one.skip(headByte);
				one.read(buf.get(), bodyByte);

				w.write(buf.get(), bodyByte);
			}
		}
		catch (...) {
			w.close();
			m_data->effect.add<Message>(L"o͂Ɏs܂");
			return false;
		}

		w.close();
		m_data->effect.add<Message>(L"o͂܂");
		return true;
	}
};

/// <summary>C֐</summary>
void Main() {
	FontAsset::Register(FSL, 128);
	FontAsset::Register(FSM, 32);
	FontAsset::Register(FSS, 16);

	MyApp myApp;
	myApp.add<MainScene>(L"MainScene");
	myApp.changeScene(L"MainScene", 0, false);
	while (System::Update()) {
		if (!myApp.updateAndDraw()) {
			break;
		}
	}
}
